# =============================================================================
# Minet CLI Utils
# =============================================================================
#
# Miscellaneous helpers used by the CLI tools.
#
import os
import sys
import stat
import yaml
import platform
from glob import iglob
from os.path import join, expanduser, isfile, relpath
from collections import namedtuple
from tqdm import tqdm
from ebbe import noop

from minet.web import (
    register_global_request_retryer_before_sleep,
    reset_global_request_retryer_before_sleep,
)
from minet.cli.exceptions import MissingColumnError
from minet.utils import fuzzy_int, prettyprint_seconds


def get_stdin_status():
    mode = os.fstat(sys.stdin.fileno()).st_mode

    if stat.S_ISFIFO(mode):
        return "piped"
    elif stat.S_ISREG(mode):
        return "redirected"

    return "terminal"


def is_stdin_empty():
    return not sys.stdin.buffer.peek(1)


def was_piped_something():
    return get_stdin_status() != "terminal" and not is_stdin_empty()


def print_err(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def die(msg=None):
    if msg is not None:
        if not isinstance(msg, list):
            msg = [msg]

        for m in msg:
            print_err(m)

    sys.exit(1)


def safe_index(l, e):
    try:
        return l.index(e)
    except ValueError:
        return None


def register_retryer_logger(print_fn=print_err):
    def callback(retry_state):
        exc = retry_state.outcome.exception()
        pretty_time = prettyprint_seconds(retry_state.idle_for, granularity=2)

        exc_name = "%s.%s" % (exc.__class__.__module__, exc.__class__.__name__)

        exc_msg = str(exc)

        msg = "\n".join(
            [
                "Failed attempt because of following exception:",
                ("%s (%s)" % (exc_name, exc_msg)) if exc_msg else exc_name,
                "Will wait for %s before attempting again." % pretty_time,
                "",
            ]
        )

        print_fn(msg)

    register_global_request_retryer_before_sleep(callback)


class LoadingBar(tqdm):
    def __init__(
        self, desc, stats=None, unit=None, unit_plural=None, total=None, **kwargs
    ):

        if unit is not None and total is None:
            if unit_plural is not None:
                unit = " " + unit_plural
            else:
                unit = " " + unit + "s"

        self.__stats = stats or {}

        if unit is not None:
            kwargs["unit"] = unit

        super().__init__(desc=desc, total=total, **kwargs)

        register_retryer_logger(self.print)

    def update_total(self, total):
        self.total = total

    def update_stats(self, **kwargs):
        for key, value in kwargs.items():
            self.__stats[key] = value

        return self.set_postfix(**self.__stats)

    def inc(self, name, amount=1):
        if name not in self.__stats:
            self.__stats[name] = 0

        self.__stats[name] += amount
        return self.update_stats()

    def print(self, *args, end="\n"):
        msg = " ".join(str(arg) for arg in args)
        self.write(msg, file=sys.stderr, end=end)

    def close(self):
        super().close()
        reset_global_request_retryer_before_sleep()

    def die(self, msg):
        self.close()
        die(msg)


def acquire_cross_platform_stdout():

    # As per #254: stdout need to be wrapped so that windows get a correct csv
    # stream output
    if "windows" in platform.system().lower():
        return open(
            sys.__stdout__.fileno(),
            mode=sys.__stdout__.mode,
            buffering=1,
            encoding=sys.__stdout__.encoding,
            errors=sys.__stdout__.errors,
            newline="",
            closefd=False,
        )

    return sys.stdout


WorkerPayload = namedtuple(
    "WorkerPayload", ["row", "headers", "path", "encoding", "content", "args"]
)


def getdefault(row, pos, default=None):
    if pos is None:
        return default

    try:
        return row[pos] or default
    except IndexError:
        return default


def create_report_iterator(cli_args, reader, worker_args=None, on_irrelevant_row=noop):
    if "filename" not in reader.headers:
        raise MissingColumnError

    filename_pos = reader.headers.filename
    error_pos = reader.headers.get("error")
    status_pos = reader.headers.get("status")
    filename_pos = reader.headers.get("filename")
    encoding_pos = reader.headers.get("encoding")
    mimetype_pos = reader.headers.get("mimetype")
    raw_content_pos = reader.headers.get("raw_contents")

    indexed_headers = reader.headers.as_dict()

    def generator():
        for i, row in enumerate(reader):
            error = getdefault(row, error_pos)

            if error is not None:
                on_irrelevant_row("errored", row, i)
                continue

            status = fuzzy_int(getdefault(row, status_pos, "200"))

            mimetype = getdefault(row, mimetype_pos, "text/html").strip()
            filename = row[filename_pos]
            encoding = getdefault(row, encoding_pos, "utf-8").strip()

            if status != 200:
                on_irrelevant_row("invalid-status", row, i)
                continue

            if not filename:
                on_irrelevant_row("no-filename", row, i)
                continue

            if "/htm" not in mimetype:
                on_irrelevant_row("invalid-mimetype", row, i)
                continue

            if raw_content_pos is not None:
                yield WorkerPayload(
                    row=row,
                    headers=indexed_headers,
                    path=None,
                    encoding=encoding,
                    content=row[raw_content_pos],
                    args=worker_args,
                )

                continue

            path = join(cli_args.input_dir or "", filename)

            yield WorkerPayload(
                row=row,
                headers=indexed_headers,
                path=path,
                encoding=encoding,
                content=None,
                args=worker_args,
            )

    return generator()


def dummy_csv_file_from_glob(pattern, root_directory=None):
    if root_directory is not None:
        pattern = join(root_directory, pattern)

    yield ["filename"]

    for p in iglob(pattern, recursive=True):
        yield [relpath(p, start=root_directory or "")]


def get_rcfile(rcfile_path=None):
    home = expanduser("~")

    files = [
        rcfile_path,
        ".minetrc",
        ".minetrc.yml",
        ".minetrc.yaml",
        ".minetrc.json",
        join(home, ".minetrc"),
        join(home, ".minetrc.yml"),
        join(home, ".minetrc.yaml"),
        join(home, ".minetrc.json"),
    ]

    for p in files:
        if p is None or not isfile(p):
            continue

        with open(p, encoding="utf-8") as f:
            return yaml.safe_load(f)

    return None



# =============================================================================
# Minet Youtube Comments CLI Action
# =============================================================================
#
# Action retrieving the comments of YouTube videos using the API.
#
import casanova

from minet.cli.utils import LoadingBar
from minet.youtube import YouTubeAPIClient
from minet.youtube.constants import YOUTUBE_COMMENT_CSV_HEADERS


def comments_action(cli_args):
    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        add=YOUTUBE_COMMENT_CSV_HEADERS,
        keep=cli_args.select,
    )

    loading_bar = LoadingBar("Retrieving comments", unit="comment", stats={"videos": 0})

    def before_sleep_until_midnight(seconds):
        loading_bar.print(
            "API limits reached. Will now wait until midnight Pacific time!"
        )

    client = YouTubeAPIClient(
        cli_args.key, before_sleep_until_midnight=before_sleep_until_midnight
    )

    for row, video in enricher.cells(cli_args.column, with_rows=True):
        generator = client.comments(video)

        for comment in generator:
            loading_bar.update()
            enricher.writerow(row, comment.as_csv_row())

        loading_bar.inc("videos")



# =============================================================================
# Minet Twitter Users CLI Action
# =============================================================================
#
# Logic of the `tw users` action.
#
import casanova
from twitwi import normalize_user, format_user_as_csv_row
from twitter import TwitterHTTPError
from twitwi.constants import USER_FIELDS, USER_PARAMS
from ebbe import as_chunks

from minet.cli.utils import LoadingBar
from minet.twitter import TwitterAPIClient
from minet.cli.twitter.utils import is_not_user_id, is_probably_not_user_screen_name


def twitter_users_action(cli_args):

    client = TwitterAPIClient(
        cli_args.access_token,
        cli_args.access_token_secret,
        cli_args.api_key,
        cli_args.api_secret_key,
        api_version="2" if cli_args.v2 else "1.1",
    )

    enricher = casanova.enricher(
        cli_args.file, cli_args.output, keep=cli_args.select, add=USER_FIELDS
    )

    loading_bar = LoadingBar(desc="Retrieving users", total=cli_args.total, unit="user")

    def call_client(users):
        kwargs = {}

        if not cli_args.v2:
            if cli_args.ids:
                kwargs["user_id"] = users
            else:
                kwargs["screen_name"] = users

            return client.call(["users", "lookup"], **kwargs)
        else:
            kwargs["params"] = USER_PARAMS
            route = ["users"]

            if cli_args.ids:
                kwargs["ids"] = users
            else:
                kwargs["usernames"] = users
                route = ["users", "by"]

            return client.call(route, **kwargs)

    for chunk in as_chunks(100, enricher.cells(cli_args.column, with_rows=True)):
        users = ",".join(row[1].lstrip("@") for row in chunk)

        for _, user in chunk:
            if cli_args.ids:
                if is_not_user_id(user):
                    loading_bar.die(
                        "The column given as argument doesn't contain user ids, you have probably given user screen names as argument instead. \nTry removing --ids from the command."
                    )

                key = "id"

            else:
                if is_probably_not_user_screen_name(user):
                    loading_bar.die(
                        "The column given as argument probably doesn't contain user screen names, you have probably given user ids as argument instead. \nTry adding --ids to the command."
                    )
                    # force flag to add

                key = "screen_name"

        try:
            result = call_client(users)
        except TwitterHTTPError as e:
            if e.e.code == 404:
                for row, user in chunk:
                    enricher.writerow(row)
            else:
                raise e

            continue

        users_raw_data = result if not cli_args.v2 else result["data"]

        indexed_result = {}

        for user in users_raw_data:
            user = normalize_user(user, v2=cli_args.v2)
            user_row = format_user_as_csv_row(user)
            indexed_result[user[key]] = user_row

        for row, user in chunk:
            user_row = indexed_result.get(user.lstrip("@"))

            enricher.writerow(row, user_row)

        loading_bar.update(len(chunk))



# =============================================================================
# Minet Web Utilities
# =============================================================================
#
# Miscellaneous web-related functions used throughout the library.
#
import re
import cgi
import certifi
import browser_cookie3
import urllib3
import ural
import json
import mimetypes
import functools
import cchardet as chardet
from collections import OrderedDict
from json.decoder import JSONDecodeError
from urllib.parse import urljoin
from urllib3 import HTTPResponse
from urllib.request import Request
from tenacity import (
    Retrying,
    wait_random_exponential,
    retry_if_exception_type,
    retry_if_exception,
    stop_after_attempt,
)

from minet.encodings import is_supported_encoding
from minet.utils import is_binary_mimetype
from minet.exceptions import (
    MaxRedirectsError,
    InfiniteRedirectsError,
    InvalidRedirectError,
    InvalidURLError,
    SelfRedirectError,
)
from minet.constants import DEFAULT_SPOOFED_UA, DEFAULT_URLLIB3_TIMEOUT, COOKIE_BROWSERS

mimetypes.init()

# Handy regexes
CHARSET_RE = re.compile(rb'<meta.*?charset=["\']*(.+?)["\'>]', flags=re.I)
PRAGMA_RE = re.compile(rb'<meta.*?content=["\']*;?charset=(.+?)["\'>]', flags=re.I)
XML_RE = re.compile(rb'^<\?xml.*?encoding=["\']*(.+?)["\'>]', flags=re.I)
NOSCRIPT_RE = re.compile(rb"<noscript[^>]*>.*</noscript[^>]*>", flags=re.I)
META_REFRESH_RE = re.compile(
    rb"""<meta\s+http-equiv=['"]?refresh['"]?\s+content=['"]?([^"']+)['">]?""",
    flags=re.I,
)
JAVASCRIPT_LOCATION_RE = re.compile(
    rb"""(?:window\.)?location(?:\s*=\s*|\.replace\(\s*)['"`](.*?)['"`]"""
)
ESCAPED_SLASH_RE = re.compile(rb"\\\/")
ASCII_RE = re.compile(r"^[ -~]*$")
HTML_RE = re.compile(
    rb"^<(?:html|head|body|title|meta|link|span|div|img|ul|ol|[ap!?])", flags=re.I
)
CANONICAL_LINK_RE = re.compile(
    rb'<link\s*[^>]*\s+rel=(?:"\s*canonical\s*"|canonical|\'\s*canonical\s*\')\s+[^>]*\s?/?>'
)
HREF_RE = re.compile(rb'href=(\"[^"]+|\'[^\']+|[^\s]+)>?\s?', flags=re.I)

# Constants
CHARDET_CONFIDENCE_THRESHOLD = 0.9
REDIRECT_STATUSES = set(HTTPResponse.REDIRECT_STATUSES)
CONTENT_CHUNK_SIZE = 1024
LARGE_CONTENT_CHUNK_SIZE = 2**16

assert CONTENT_CHUNK_SIZE < LARGE_CONTENT_CHUNK_SIZE


def prebuffer_response_up_to(response, target):
    try:
        if response._body is None:
            response._body = response.read(target)
        else:
            target -= len(response._body)

            if target <= 0:
                return None

            response._body += response.read(target)

    except Exception as e:
        return e

    return None


# TODO: add a version that tallies the possibilities
def guess_response_encoding(response, is_xml=False, use_chardet=False):
    """
    Function taking an urllib3 response object and attempting to guess its
    encoding.
    """
    content_type_header = response.getheader("content-type")

    suboptimal_charset = None

    if content_type_header is not None:
        parsed_header = cgi.parse_header(content_type_header)

        if len(parsed_header) > 1:
            charset = parsed_header[1].get("charset")

            if charset is not None:
                if is_supported_encoding(charset):
                    return charset.lower()
                else:
                    suboptimal_charset = charset

    data = response.data
    chunk = data[:CONTENT_CHUNK_SIZE]

    # Data is empty
    if not chunk.strip():
        return None

    # TODO: use re.search to go faster!
    if is_xml:
        matches = re.findall(CHARSET_RE, chunk)

        if len(matches) == 0:
            matches = re.findall(PRAGMA_RE, chunk)

        if len(matches) == 0:
            matches = re.findall(XML_RE, chunk)

        # NOTE: here we are returning the last one, but we could also use
        # frequency at the expense of performance
        if len(matches) != 0:
            charset = matches[-1].lower().decode()

            if is_supported_encoding(charset):
                return charset
            else:
                suboptimal_charset = charset

    if use_chardet:
        chardet_result = chardet.detect(data)

        # Could not detect anything
        if not chardet_result or chardet_result.get("confidence") is None:
            return None

        if chardet_result["confidence"] >= CHARDET_CONFIDENCE_THRESHOLD:
            return chardet_result["encoding"].lower()

    return suboptimal_charset


def looks_like_html(html_chunk):
    return HTML_RE.match(html_chunk) is not None


def parse_http_header(header):
    key, value = header.split(":", 1)

    return key.strip(), value.strip()


# TODO: take more cases into account...
#   http://www.otsukare.info/2015/03/26/refresh-http-header
def parse_http_refresh(value):
    try:

        if isinstance(value, bytes):
            value = value.decode()

        duration, url = value.strip().split(";", 1)

        if not url.lower().strip().startswith("url="):
            return None

        return int(duration), str(url.split("=", 1)[1])
    except Exception:
        return None


def extract_href(value):

    m = HREF_RE.search(value)

    if not m:
        return None

    url = m.group(1)

    try:
        url = url.decode("utf-8")
    except UnicodeDecodeError:
        return None

    return url.strip("\"'") or None


def find_canonical_link(html_chunk):
    m = CANONICAL_LINK_RE.search(html_chunk)

    if not m:
        return None

    return extract_href(m.group())


def find_meta_refresh(html_chunk):
    m = META_REFRESH_RE.search(html_chunk)

    if not m:
        return None

    return parse_http_refresh(m.group(1))


def find_javascript_relocation(html_chunk):
    m = JAVASCRIPT_LOCATION_RE.search(html_chunk)

    if not m:
        return None

    try:
        return ESCAPED_SLASH_RE.sub(b"/", m.group(1)).decode()
    except Exception:
        return None


class CookieResolver(object):
    def __init__(self, jar):
        self.jar = jar

    def __call__(self, url):
        req = Request(url)
        self.jar.add_cookie_header(req)

        return req.get_header("Cookie") or None


def grab_cookies(browser="firefox"):
    if browser not in COOKIE_BROWSERS:
        raise TypeError('minet.utils.grab_cookies: unknown "%s" browser.' % browser)

    try:
        return CookieResolver(getattr(browser_cookie3, browser)())
    except browser_cookie3.BrowserCookieError:
        return None


def dict_to_cookie_string(d):
    return "; ".join("%s=%s" % r for r in d.items())


def create_pool(proxy=None, threads=None, insecure=False, **kwargs):
    """
    Helper function returning a urllib3 pool manager with sane defaults.
    """

    manager_kwargs = {"timeout": DEFAULT_URLLIB3_TIMEOUT}

    if not insecure:
        manager_kwargs["cert_reqs"] = "CERT_REQUIRED"
        manager_kwargs["ca_certs"] = certifi.where()
    else:
        manager_kwargs["cert_reqs"] = "CERT_NONE"
        # manager_kwargs['assert_hostname'] = False

        urllib3.disable_warnings()

    if threads is not None:

        # TODO: maxsize should increase with group_parallelism
        manager_kwargs["maxsize"] = 10
        manager_kwargs["num_pools"] = threads * 2

    manager_kwargs.update(kwargs)

    if proxy is not None:
        return urllib3.ProxyManager(proxy, **manager_kwargs)

    return urllib3.PoolManager(**manager_kwargs)


DEFAULT_POOL = create_pool(maxsize=10, num_pools=10)


def raw_request(
    http,
    url,
    method="GET",
    headers=None,
    preload_content=True,
    release_conn=True,
    timeout=None,
    body=None,
):
    """
    Generic request helpers using a urllib3 pool to access some resource.
    """

    # Validating URL
    if not ural.is_url(
        url, require_protocol=True, tld_aware=True, allow_spaces_in_path=True
    ):
        return InvalidURLError(url=url), None

    # Performing request
    request_kwargs = {
        "headers": headers,
        "body": body,
        "preload_content": preload_content,
        "release_conn": release_conn,
        "redirect": False,
        "retries": False,
    }

    if timeout is not None:
        request_kwargs["timeout"] = timeout

    try:
        response = http.request(method, url, **request_kwargs)
    except Exception as e:
        return e, None

    return None, response


class Redirection(object):
    __slots__ = ("status", "type", "url")

    def __init__(self, url, _type="hit"):
        self.status = None
        self.url = url
        self.type = _type

    def __repr__(self):
        class_name = self.__class__.__name__

        return ("<%(class_name)s type=%(type)s status=%(status)s url=%(url)s>") % {
            "class_name": class_name,
            "type": self.type,
            "status": self.status,
            "url": self.url,
        }


def raw_resolve(
    http,
    url,
    method="GET",
    headers=None,
    max_redirects=5,
    follow_refresh_header=True,
    follow_meta_refresh=False,
    follow_js_relocation=False,
    return_response=False,
    infer_redirection=False,
    timeout=None,
    body=None,
    canonicalize=False,
):
    """
    Helper function attempting to resolve the given url.
    """

    url_stack = OrderedDict()
    error = None
    response = None

    for _ in range(max_redirects):
        if infer_redirection:
            target = ural.infer_redirection(url, recursive=False)

            if target != url:
                url_stack[url] = Redirection(url, "infer")
                url = target
                continue

        if response:
            response.release_conn()
            response.close()

        http_error, response = raw_request(
            http,
            url,
            method=method,
            headers=headers,
            body=body,
            preload_content=False,
            release_conn=False,
            timeout=timeout,
        )

        redirection = Redirection(url)

        # Request error
        if http_error:
            redirection.type = "error"
            url_stack[url] = redirection
            error = http_error
            break

        # Cycle
        if url in url_stack:
            error = InfiniteRedirectsError("Infinite redirects")
            break

        redirection.status = response.status
        url_stack[url] = redirection
        location = None

        if response.status not in REDIRECT_STATUSES:

            if response.status < 400:

                if follow_refresh_header:
                    refresh = response.getheader("refresh")

                    if refresh is not None:
                        p = parse_http_refresh(refresh)

                        if p is not None:
                            location = p[1]
                            redirection.type = "refresh-header"

                # Reading a small chunk of the html
                if location is None and (follow_meta_refresh or follow_js_relocation):
                    error = prebuffer_response_up_to(response, CONTENT_CHUNK_SIZE)

                    if error is not None:
                        redirection.type = "error"
                        break

                # Meta refresh
                if location is None and follow_meta_refresh:
                    meta_refresh = find_meta_refresh(response._body)

                    if meta_refresh is not None:
                        location = meta_refresh[1]
                        redirection.type = "meta-refresh"

                # JavaScript relocation
                if location is None and follow_js_relocation:
                    js_relocation = find_javascript_relocation(response._body)

                    if js_relocation is not None:
                        location = js_relocation
                        redirection.type = "js-relocation"

            # Found the end
            if location is None:
                redirection.type = "hit"

            # Canonical url
            if redirection.type == "hit":
                if canonicalize:
                    error = prebuffer_response_up_to(response, LARGE_CONTENT_CHUNK_SIZE)

                    if error is not None:
                        redirection.type = "error"
                        break

                    canonical = find_canonical_link(response._body)

                    if canonical is not None and canonical != url:
                        canonical = urljoin(url, canonical)
                        redirection = Redirection(canonical, "canonical")
                        url_stack[canonical] = redirection

                break

        else:
            redirection.type = "location-header"
            location = response.getheader("location")

        # Invalid redirection
        if not location:
            error = InvalidRedirectError("Redirection is invalid")
            break

        # Location badly encoded?
        try:
            if not ASCII_RE.match(location):
                byte_location = location.encode("latin1")
                detection = chardet.detect(byte_location)
                guessed_encoding = detection["encoding"].lower()

                if (
                    guessed_encoding != "iso-8859-1"
                    and guessed_encoding != "ascii"
                    and detection["confidence"] >= CHARDET_CONFIDENCE_THRESHOLD
                ):
                    location = byte_location.decode(guessed_encoding)
        except Exception:
            pass

        # Resolving next url
        next_url = urljoin(url, location.strip())

        # Self loop?
        if next_url == url:
            error = SelfRedirectError("Self redirection")
            break

        # Go to next
        url = next_url

    # We reached max redirects
    else:
        error = MaxRedirectsError("Maximum number of redirects exceeded")

    if response and not return_response:
        response.release_conn()
        response.close()

    compiled_stack = list(url_stack.values())

    if return_response:
        return error, compiled_stack, response

    return error, compiled_stack


def build_request_headers(headers=None, cookie=None, spoof_ua=False, json_body=False):

    # Formatting headers
    final_headers = {}

    if spoof_ua:
        final_headers["User-Agent"] = DEFAULT_SPOOFED_UA

    if cookie:
        if not isinstance(cookie, str):
            cookie = dict_to_cookie_string(cookie)

        final_headers["Cookie"] = cookie

    if json_body:
        final_headers["Content-Type"] = "application/json"

    # Note: headers passed explicitly by users always win
    if headers is not None:
        final_headers.update(headers)

    return final_headers


def request(
    url,
    pool=DEFAULT_POOL,
    method="GET",
    headers=None,
    cookie=None,
    spoof_ua=True,
    follow_redirects=True,
    max_redirects=5,
    follow_refresh_header=True,
    follow_meta_refresh=False,
    follow_js_relocation=False,
    timeout=None,
    body=None,
    json_body=None,
):

    # Formatting headers
    final_headers = build_request_headers(
        headers=headers,
        cookie=cookie,
        spoof_ua=spoof_ua,
        json_body=json_body is not None,
    )

    # Dealing with body
    final_body = None

    if isinstance(body, bytes):
        final_body = body
    elif isinstance(body, str):
        final_body = body.encode("utf-8")

    if json_body is not None:
        final_body = json.dumps(json_body, ensure_ascii=False).encode("utf-8")

    if not follow_redirects:
        return raw_request(
            pool, url, method, headers=final_headers, body=final_body, timeout=timeout
        )
    else:
        err, _, response = raw_resolve(
            pool,
            url,
            method,
            headers=final_headers,
            body=final_body,
            max_redirects=max_redirects,
            return_response=True,
            follow_refresh_header=follow_refresh_header,
            follow_meta_refresh=follow_meta_refresh,
            follow_js_relocation=follow_js_relocation,
            timeout=timeout,
        )

        if err:
            return err, response

        # Finishing reading body
        try:
            response._body = (response._body or b"") + response.read()
        except Exception as e:
            return e, response
        finally:
            if response is not None:
                response.close()
                response.release_conn()

        return None, response


def resolve(
    url,
    pool=DEFAULT_POOL,
    method="GET",
    headers=None,
    cookie=None,
    spoof_ua=True,
    max_redirects=5,
    follow_refresh_header=True,
    follow_meta_refresh=False,
    follow_js_relocation=False,
    infer_redirection=False,
    timeout=None,
    canonicalize=False,
):

    final_headers = build_request_headers(
        headers=headers, cookie=cookie, spoof_ua=spoof_ua
    )

    return raw_resolve(
        pool,
        url,
        method,
        headers=final_headers,
        max_redirects=max_redirects,
        follow_refresh_header=follow_refresh_header,
        follow_meta_refresh=follow_meta_refresh,
        follow_js_relocation=follow_js_relocation,
        infer_redirection=infer_redirection,
        timeout=timeout,
        canonicalize=canonicalize,
    )


def extract_response_meta(response, guess_encoding=True, guess_extension=True):
    meta = {"ext": None, "mimetype": None, "encoding": None, "is_text": None}

    # Guessing extension
    if guess_extension:

        # Guessing mime type
        # TODO: validate mime type string?
        mimetype, _ = mimetypes.guess_type(response.geturl())

        if "Content-Type" in response.headers:
            content_type = response.headers["Content-Type"]
            parsed_header = cgi.parse_header(content_type)

            if parsed_header and parsed_header[0].strip():
                mimetype = parsed_header[0].strip()

        if mimetype is None and looks_like_html(response.data[:CONTENT_CHUNK_SIZE]):
            mimetype = "text/html"

        if mimetype is not None:
            ext = mimetypes.guess_extension(mimetype)

            if ext == ".htm":
                ext = ".html"
            elif ext == ".jpe":
                ext = ".jpg"

            meta["mimetype"] = mimetype
            meta["ext"] = ext

    if meta["mimetype"] is not None:
        meta["is_text"] = not is_binary_mimetype(meta["mimetype"])

        if not meta["is_text"]:
            guess_encoding = False

    # Guessing encoding
    if guess_encoding:
        meta["encoding"] = guess_response_encoding(
            response, is_xml=True, use_chardet=True
        )

    return meta


def request_jsonrpc(url, method, pool=DEFAULT_POOL, *args, **kwargs):
    params = []

    if len(args) > 0:
        params = args
    elif len(kwargs) > 0:
        params = kwargs

    err, response = request(
        url, pool=pool, method="POST", json_body={"method": method, "params": params}
    )

    if err is not None:
        return err, None

    data = json.loads(response.data)

    return None, data


def request_json(url, pool=DEFAULT_POOL, *args, **kwargs):
    err, response = request(url, pool=pool, *args, **kwargs)

    if err:
        return err, response, None

    try:
        return None, response, json.loads(response.data.decode())
    except (JSONDecodeError, UnicodeDecodeError) as e:
        return e, response, None


def request_text(url, pool=DEFAULT_POOL, *args, encoding="utf-8", **kwargs):
    err, response = request(url, pool=pool, *args, **kwargs)

    if err:
        return err, response, None

    return None, response, response.data.decode(encoding)


THREE_HOURS = 3 * 60 * 60
GLOBAL_RETRYER_BEFORE_SLEEP = None


def register_global_request_retryer_before_sleep(callback):
    global GLOBAL_RETRYER_BEFORE_SLEEP

    if not callable(callback):
        raise TypeError

    GLOBAL_RETRYER_BEFORE_SLEEP = callback


def reset_global_request_retryer_before_sleep():
    global GLOBAL_RETRYER_BEFORE_SLEEP

    GLOBAL_RETRYER_BEFORE_SLEEP = None


def wrap_before_sleep_callback_with_global_hook(callback):
    def chain(*args, **kwargs):
        if GLOBAL_RETRYER_BEFORE_SLEEP is not None:
            GLOBAL_RETRYER_BEFORE_SLEEP(*args, **kwargs)

        if callback is not None:
            callback(*args, **kwargs)

    return chain


def create_request_retryer(
    min=10,
    max=THREE_HOURS,
    max_attempts=9,
    before_sleep=None,
    additional_exceptions=None,
    predicate=None,
):
    global GLOBAL_RETRYER_BEFORE_SLEEP

    retryable_exception_types = [
        urllib3.exceptions.TimeoutError,
        urllib3.exceptions.ProtocolError,
    ]

    if additional_exceptions:
        for exc in additional_exceptions:
            retryable_exception_types.append(exc)

    before_sleep_chain = wrap_before_sleep_callback_with_global_hook(before_sleep)

    retry_condition = retry_if_exception_type(
        exception_types=tuple(retryable_exception_types)
    )

    if callable(predicate):
        retry_condition |= retry_if_exception(predicate)

    return Retrying(
        wait=wait_random_exponential(exp_base=6, min=min, max=max),
        retry=retry_condition,
        stop=stop_after_attempt(max_attempts),
        before_sleep=before_sleep_chain,
    )


def retrying_method(attr="retryer"):
    def decorate(fn):
        @functools.wraps(fn)
        def decorated(self, *args, **kwargs):
            retryer = getattr(self, attr)

            if not isinstance(retryer, Retrying):
                raise ValueError

            return retryer(fn, self, *args, **kwargs)

        return decorated

    return decorate



# =============================================================================
# Minet Url Parse CLI Action
# =============================================================================
#
# Logic of the `url-parse` action.
#
import casanova
from ural import (
    is_url,
    is_shortened_url,
    normalize_url,
    get_hostname,
    get_domain_name,
    get_normalized_hostname,
    infer_redirection,
    is_typo_url,
)
from ural.facebook import (
    parse_facebook_url,
    FacebookPost,
    FacebookUser,
    FacebookGroup,
    FacebookHandle,
    FacebookPhoto,
    FacebookVideo,
)
from ural.youtube import parse_youtube_url, YoutubeVideo, YoutubeUser, YoutubeChannel
from ural.twitter import parse_twitter_url, TwitterTweet, TwitterUser

from minet.cli.utils import LoadingBar

REPORT_HEADERS = [
    "normalized_url",
    "inferred_redirection",
    "domain_name",
    "hostname",
    "normalized_hostname",
    "probably_shortened",
    "probably_typo",
]

FACEBOOK_REPORT_HEADERS = [
    "facebook_type",
    "facebook_id",
    "facebook_full_id",
    "facebook_handle",
    "facebook_normalized_url",
]

YOUTUBE_REPORT_HEADERS = ["youtube_type", "youtube_id", "youtube_name"]

TWITTER_REPORT_HEADERS = ["twitter_type", "twitter_user_screen_name", "tweet_id"]


def extract_standard_addendum(cli_args, url):
    inferred_redirection = infer_redirection(url)

    return [
        normalize_url(
            url, strip_protocol=cli_args.strip_protocol, strip_trailing_slash=True
        ),
        inferred_redirection if inferred_redirection != url else "",
        get_domain_name(url),
        get_hostname(url),
        get_normalized_hostname(url),
        "yes" if is_shortened_url(url) else "",
        "yes" if is_typo_url(url) else "",
    ]


YOUTUBE_TYPES = {YoutubeVideo: "video", YoutubeUser: "user", YoutubeChannel: "channel"}


def extract_youtube_addendum(url):
    parsed = parse_youtube_url(url)

    if parsed is None:
        return None

    return [YOUTUBE_TYPES.get(type(parsed)), parsed.id, getattr(parsed, "name", "")]


def extract_facebook_addendum(url):
    parsed = parse_facebook_url(url)

    if parsed is None:
        return None

    if isinstance(parsed, FacebookPost):
        return ["post", parsed.id, parsed.full_id or "", "", parsed.url]

    elif isinstance(parsed, FacebookHandle):
        return ["handle", "", "", parsed.handle, parsed.url]

    elif isinstance(parsed, FacebookUser):
        return ["user", parsed.id or "", "", parsed.handle or "", parsed.url]

    elif isinstance(parsed, FacebookGroup):
        return ["group", parsed.id or "", "", parsed.handle or "", parsed.url]

    elif isinstance(parsed, FacebookPhoto):
        return ["photo", parsed.id, "", "", parsed.url]

    elif isinstance(parsed, FacebookVideo):
        return ["video", parsed.id, "", "", parsed.url]

    else:
        raise TypeError("unknown facebook parse result type!")


def extract_twitter_addendum(url):
    parsed = parse_twitter_url(url)

    if parsed is None:
        return None

    if isinstance(parsed, TwitterUser):
        return ["user", parsed.screen_name, ""]

    elif isinstance(parsed, TwitterTweet):
        return ["tweet", parsed.user_screen_name, parsed.id]

    else:
        raise TypeError("unknown twitter parse result type!")


def url_parse_action(cli_args):
    headers = REPORT_HEADERS

    if cli_args.facebook:
        headers = FACEBOOK_REPORT_HEADERS
    elif cli_args.youtube:
        headers = YOUTUBE_REPORT_HEADERS
    elif cli_args.twitter:
        headers = TWITTER_REPORT_HEADERS

    multiplex = None

    if cli_args.separator is not None:
        multiplex = (cli_args.column, cli_args.separator)

    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        add=headers,
        keep=cli_args.select,
        multiplex=multiplex,
    )

    loading_bar = LoadingBar(desc="Parsing", unit="row", total=cli_args.total)

    for row, url in enricher.cells(cli_args.column, with_rows=True):
        loading_bar.update()

        url = url.strip()

        if not is_url(url, allow_spaces_in_path=True, require_protocol=False):
            enricher.writerow(row)
            continue

        if cli_args.facebook:
            addendum = extract_facebook_addendum(url)
        elif cli_args.youtube:
            addendum = extract_youtube_addendum(url)
        elif cli_args.twitter:
            addendum = extract_twitter_addendum(url)
        else:
            addendum = extract_standard_addendum(cli_args, url)

        if addendum is None:
            enricher.writerow(row)
            continue

        enricher.writerow(row, addendum)



# =============================================================================
# Minet Twitter Friends CLI Action
# =============================================================================
#
# Logic of the `tw friends` action.
#
from minet.cli.twitter.utils import make_twitter_action

REPORT_HEADERS = ["friend_id"]

twitter_friends_action = make_twitter_action(
    method_name="friends", csv_headers=REPORT_HEADERS
)



# =============================================================================
# Minet CrowdTangle CLI Action
# =============================================================================
#
# Logic of the `ct` action.
#
from minet.cli.utils import die


def crowdtangle_action(cli_args):

    # A token is needed to be able to access the API
    if not cli_args.token:
        die(
            [
                "A token is needed to be able to access CrowdTangle's API.",
                "You can provide one using the `--token` argument.",
            ]
        )

    if cli_args.ct_action == "posts":
        from minet.cli.crowdtangle.posts import crowdtangle_posts_action

        crowdtangle_posts_action(cli_args)

    elif cli_args.ct_action == "posts-by-id":
        from minet.cli.crowdtangle.posts_by_id import crowdtangle_posts_by_id_action

        crowdtangle_posts_by_id_action(cli_args)

    elif cli_args.ct_action == "lists":
        from minet.cli.crowdtangle.lists import crowdtangle_lists_action

        crowdtangle_lists_action(cli_args)

    elif cli_args.ct_action == "leaderboard":
        from minet.cli.crowdtangle.leaderboard import crowdtangle_leaderboard_action

        crowdtangle_leaderboard_action(cli_args)

    elif cli_args.ct_action == "search":
        from minet.cli.crowdtangle.search import crowdtangle_search_action

        crowdtangle_search_action(cli_args)

    elif cli_args.ct_action == "summary":
        from minet.cli.crowdtangle.summary import crowdtangle_summary_action

        crowdtangle_summary_action(cli_args)



# =============================================================================
# Minet BuzzSumo Constants
# =============================================================================
#
# General constants used by the buzzsumo package.
#
ARTICLES_CSV_HEADERS = [
    "id",
    "published_date",
    "updated_at",
    "url",
    "og_url",
    "thumbnail",
    "domain_name",
    "subdomain",
    "num_linking_domains",
    "total_shares",
    "total_facebook_shares",
    "facebook_comments",
    "facebook_shares",
    "facebook_likes",
    "love_count",
    "wow_count",
    "haha_count",
    "angry_count",
    "sad_count",
    "twitter_shares",
    "twitter_user_id",
    "pinterest_shares",
    "total_reddit_engagements",
    "alexa_rank",
    "evergreen_score",
    "evergreen_score2",
    "youtube_trending_score",
    "youtube_views",
    "youtube_likes",
    "youtube_dislikes",
    "youtube_comments",
    "article_amplifiers",
    "title",
    "display_title",
    "author_name",
    "language",
    "num_words",
    "article_types",
    "video_length",
]



# =============================================================================
# Minet Hyphe Declare CLI Action
# =============================================================================
#
# Logic of the `hyphe declare` action.
#
import casanova
from urllib.parse import unquote

from minet.cli.utils import die, LoadingBar
from minet.cli.exceptions import InvalidArgumentsError
from minet.hyphe import HypheAPIClient
from minet.hyphe.exceptions import HypheCorpusAuthenticationError, HypheRequestFailError


def extract_tags(row, tag_pos_list):
    tags = {}

    for name, pos in tag_pos_list:
        cell_value = row[pos]

        if not cell_value:
            continue

        # TODO: hyphe tag serialization is not bijective
        tags[name] = cell_value.split("|")

    return {"USER": tags}


def hyphe_declare_action(cli_args):
    reader = casanova.reader(cli_args.webentities, total=cli_args.total)
    headers = reader.headers

    name_pos = headers.get("NAME")
    homepage_pos = headers.get("HOME PAGE")
    prefixes_pos = headers.get("PREFIXES AS LRU")
    status_pos = headers.get("STATUS")
    startpages_pos = headers.get("START PAGES")

    tag_pos_list = [
        (name.split("(TAGS)", 1)[0].strip(), pos)
        for name, pos in headers
        if name.endswith("(TAGS)")
    ]

    if (
        name_pos is None
        or prefixes_pos is None
        or homepage_pos is None
        or status_pos is None
        or startpages_pos is None
    ):
        raise InvalidArgumentsError(
            "input csv file is not a valid hyphe webentities export"
        )

    client = HypheAPIClient(cli_args.url)
    corpus = client.corpus(cli_args.corpus, password=cli_args.password)

    try:
        corpus.ensure_is_started()
    except HypheCorpusAuthenticationError:
        die(
            [
                'Wrong password for the "%s" corpus!' % cli_args.corpus,
                "Don't forget to provide a password for this corpus using --password",
            ]
        )

    loading_bar = LoadingBar(
        desc="Declaring web entities",
        unit="webentity",
        unit_plural="webentities",
        total=reader.total,
    )

    for row in reader:
        loading_bar.update()

        startpages_cell = row[startpages_pos]

        name = row[name_pos]

        # TODO: bug on hyphe's side
        if "h:localhost|" in row[prefixes_pos]:
            loading_bar.print("Dropping %s because of localhost issue" % name)
            continue

        homepage = row[homepage_pos]
        prefixes = row[prefixes_pos].split(" ")
        startpages = startpages_cell.split(" ") if startpages_cell else []
        startpages = list(set(startpages))  # TODO: dedupe because of hyphe issues
        tags = extract_tags(row, tag_pos_list)

        # 1. Declaring the entities
        err, result = corpus.call(
            "store.declare_webentity_by_lrus",
            list_lrus=prefixes,
            name=name,
            status=row[status_pos],
            lruVariations=False,
            startpages=startpages,
            tags=tags,
        )

        if err:
            raise err

        webentity_id = result["result"]["id"]

        # 2. Setting the homepage
        try:
            err, _ = corpus.call(
                "store.set_webentity_homepage",
                webentity_id=webentity_id,
                homepage=unquote(
                    homepage
                ),  # TODO: drop unquote, linked to hyphe issue #447
            )
        except HypheRequestFailError as e:
            if "belong" in str(e):
                loading_bar.print("Homepage error with %s %s" % (name, homepage))
            else:
                raise

        if err:
            raise err



# =============================================================================
# Minet BuzzSumo CLI Action
# =============================================================================
#
# Logic of the `bz` action.
#
from minet.cli.utils import die


def buzzsumo_action(cli_args):

    # A token is needed to be able to access the API
    if not cli_args.token:
        die(
            [
                "A token is needed to be able to access BuzzSumo's API.",
                "You can provide one using the `--token` argument.",
            ]
        )

    if cli_args.bz_action == "limit":
        from minet.cli.buzzsumo.limit import buzzsumo_limit_action

        buzzsumo_limit_action(cli_args)

    if cli_args.bz_action == "domain-summary":
        from minet.cli.buzzsumo.domain_summary import buzzsumo_domain_summary_action

        buzzsumo_domain_summary_action(cli_args)

    if cli_args.bz_action == "domain":
        from minet.cli.buzzsumo.domain import buzzsumo_domain_action

        buzzsumo_domain_action(cli_args)



# =============================================================================
# Minet Google Sheets CLI Action
# =============================================================================
#
# Logic of the `google sheets` action.
#
from browser_cookie3 import BrowserCookieError

from minet.cli.utils import die
from minet.google import export_google_sheets_as_csv
from minet.google.exceptions import (
    GoogleSheetsInvalidTargetError,
    GoogleSheetsInvalidContentTypeError,
    GoogleSheetsMissingCookieError,
    GoogleSheetsNotFoundError,
    GoogleSheetsUnauthorizedError,
    GoogleSheetsMaxAttemptsExceeded,
)


def google_sheets_action(cli_args):
    try:
        data = export_google_sheets_as_csv(
            cli_args.url, cookie=cli_args.cookie, authuser=cli_args.authuser
        )
    except GoogleSheetsInvalidTargetError:
        die("Could not extract a valid google sheets id from provided argument!")
    except BrowserCookieError:
        die("Could not extract cookie from %s!" % cli_args.cookie)
    except GoogleSheetsMissingCookieError:
        die("Did not find a relevant cookie!")
    except GoogleSheetsInvalidContentTypeError:
        die("Could not export spreadsheet as CSV!")
    except GoogleSheetsNotFoundError:
        die("Could not find spreadsheet (404)!")
    except GoogleSheetsUnauthorizedError:
        die(
            "You don't have access to this spreadsheet. Did you forget to set --cookie?"
        )
    except GoogleSheetsMaxAttemptsExceeded:
        die(
            "Maximum number of attempts exceeded! You can still set --authuser if you logged in numerous google accounts at once."
        )

    cli_args.output.write(data)



# =============================================================================
# Minet BuzzSumo Domain CLI Action
# =============================================================================
#
# Logic of the `bz domain` action.
#
import casanova

from minet.cli.utils import LoadingBar
from minet.buzzsumo import BuzzSumoAPIClient
from minet.buzzsumo.constants import ARTICLES_CSV_HEADERS
from minet.buzzsumo.exceptions import (
    BuzzSumoInvalidQueryError,
    BuzzSumoInvalidTokenError,
)


def buzzsumo_domain_action(cli_args):

    client = BuzzSumoAPIClient(cli_args.token)

    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        keep=cli_args.select,
        add=ARTICLES_CSV_HEADERS,
    )

    loading_bar = LoadingBar(desc="Retrieving articles", unit="article")

    for row, domain_name in enricher.cells(cli_args.column, with_rows=True):
        loading_bar.print('Retrieving articles for "%s"' % domain_name)
        try:
            for article in client.domain_articles(
                domain_name, cli_args.begin_date, cli_args.end_date
            ):
                loading_bar.update()
                enricher.writerow(row, article.as_csv_row())
        except BuzzSumoInvalidTokenError:
            loading_bar.die("Your BuzzSumo token is invalid!")
        except BuzzSumoInvalidQueryError as e:
            loading_bar.die(
                "Invalid query: %s" % e.url + "\nMessage from the API: %s" % e
            )

        loading_bar.inc("domains")



# =============================================================================
# Minet Mediacloud CLI Action
# =============================================================================
#
# Logic of the `mc` action.
#
from minet.cli.utils import die


def mediacloud_action(cli_args):

    # A token is needed to be able to access the API
    if not cli_args.token:
        die(
            [
                "A token is needed to be able to access Mediacloud's API.",
                "You can provide one using the `--token` argument.",
            ]
        )

    if cli_args.mc_action == "medias":
        from minet.cli.mediacloud.medias import mediacloud_medias_action

        mediacloud_medias_action(cli_args)

    if cli_args.mc_action == "topic":
        from minet.cli.mediacloud.topic import mediacloud_topic_action

        mediacloud_topic_action(cli_args)

    elif cli_args.mc_action == "search":
        from minet.cli.mediacloud.search import mediacloud_search_action

        mediacloud_search_action(cli_args)



# =============================================================================
# Minet Instagram Exceptions
# =============================================================================
#
from minet.exceptions import MinetError


class InstagramError(MinetError):
    pass


class InstagramPublicAPIInvalidResponseError(InstagramError):
    pass


class InstagramInvalidCookieError(InstagramError):
    pass



# =============================================================================
# Minet CrowdTangle API Client
# =============================================================================
#
# A unified CrowdTangle API client that can be used to keep an eye on the
# rate limit and the used token etc.
#
import json

from minet.utils import RateLimiterState, rate_limited_method
from minet.web import create_pool, create_request_retryer, retrying_method, request
from minet.crowdtangle.constants import (
    CROWDTANGLE_DEFAULT_TIMEOUT,
    CROWDTANGLE_DEFAULT_RATE_LIMIT,
    CROWDTANGLE_LINKS_DEFAULT_RATE_LIMIT,
)
from minet.crowdtangle.exceptions import (
    CrowdTangleInvalidJSONError,
    CrowdTangleInvalidTokenError,
    CrowdTangleRateLimitExceeded,
    CrowdTangleInvalidRequestError,
    CrowdTangleServerError,
)
from minet.crowdtangle.leaderboard import crowdtangle_leaderboard
from minet.crowdtangle.lists import crowdtangle_lists
from minet.crowdtangle.post import crowdtangle_post
from minet.crowdtangle.posts import crowdtangle_posts
from minet.crowdtangle.search import crowdtangle_search
from minet.crowdtangle.summary import crowdtangle_summary


class CrowdTangleAPIClient(object):
    def __init__(self, token, rate_limit=None, before_sleep=None):
        if rate_limit is None:
            rate_limit = CROWDTANGLE_DEFAULT_RATE_LIMIT
            summary_rate_limit = CROWDTANGLE_LINKS_DEFAULT_RATE_LIMIT
        else:
            rate_limit = rate_limit
            summary_rate_limit = rate_limit

        self.token = token
        self.rate_limiter_state = RateLimiterState(rate_limit, period=60)
        self.summary_rate_limiter_state = RateLimiterState(
            summary_rate_limit, period=60
        )
        self.pool = create_pool(timeout=CROWDTANGLE_DEFAULT_TIMEOUT)
        self.retryer = create_request_retryer(
            additional_exceptions=[CrowdTangleInvalidJSONError, CrowdTangleServerError],
            before_sleep=before_sleep,
        )

    @retrying_method()
    def __request(self, url):
        err, response = request(url, pool=self.pool)

        # Debug
        if err:
            raise err

        # Bad auth
        if response.status == 401:
            raise CrowdTangleInvalidTokenError

        # Rate limited
        if response.status == 429:
            raise CrowdTangleRateLimitExceeded

        # Server error
        if response.status >= 500:
            raise CrowdTangleServerError(url=url, status=response.status)

        # Bad params
        if response.status >= 400:
            data = response.data.decode("utf-8")

            try:
                data = json.loads(data)
            except json.decoder.JSONDecodeError:
                raise CrowdTangleInvalidRequestError(
                    data, url=url, status=response.status
                )

            raise CrowdTangleInvalidRequestError(
                data["message"], url=url, code=data.get("code"), status=response.status
            )

        try:
            data = json.loads(response.data)["result"]
        except (json.decoder.JSONDecodeError, TypeError, KeyError):
            raise CrowdTangleInvalidJSONError

        return data

    @rate_limited_method("rate_limiter_state")
    def request(self, url):
        return self.__request(url)

    @rate_limited_method("summary_rate_limiter_state")
    def request_summary(self, url):
        return self.__request(url)

    def leaderboard(self, **kwargs):
        return crowdtangle_leaderboard(self.request, token=self.token, **kwargs)

    def lists(self, **kwargs):
        return crowdtangle_lists(self.request, token=self.token, **kwargs)

    def post(self, post_id, **kwargs):
        return crowdtangle_post(self.request, post_id, token=self.token, **kwargs)

    def posts(self, sort_by="date", **kwargs):
        return crowdtangle_posts(
            self.request, token=self.token, sort_by=sort_by, **kwargs
        )

    def search(self, terms, sort_by="date", **kwargs):
        return crowdtangle_search(
            self.request, token=self.token, terms=terms, sort_by=sort_by, **kwargs
        )

    def summary(self, link, **kwargs):
        return crowdtangle_summary(
            self.request_summary, link, token=self.token, **kwargs
        )



# =============================================================================
# Minet Twitter Scrape CLI Action
# =============================================================================
#
# Logic of the `tw scrape` action.
#
import casanova
from twitwi.constants import TWEET_FIELDS, USER_FIELDS
from twitwi import format_tweet_as_csv_row, format_user_as_csv_row

from minet.utils import PseudoFStringFormatter
from minet.cli.utils import LoadingBar
from minet.twitter import TwitterAPIScraper
from minet.twitter.exceptions import TwitterPublicAPIOverCapacityError
from minet.twitter.constants import ADDITIONAL_TWEET_FIELDS

CUSTOM_FORMATTER = PseudoFStringFormatter()


def format_meta_row(meta):
    if meta is None:
        return [""] * len(ADDITIONAL_TWEET_FIELDS)

    return [
        meta.get("intervention_type"),
        meta.get("intervention_text"),
        meta.get("intervention_url"),
    ]


def twitter_scrape_action(cli_args):
    scraper = TwitterAPIScraper()

    unit = cli_args.items[:-1]

    # Stats
    loading_bar = LoadingBar(
        "Collecting %s" % unit,
        total=cli_args.limit,
        unit=unit,
        stats={"queries": 0},
    )

    headers = (
        (TWEET_FIELDS + ADDITIONAL_TWEET_FIELDS)
        if cli_args.items == "tweets"
        else USER_FIELDS
    )

    enricher = casanova.enricher(
        cli_args.file, cli_args.output, add=headers, keep=cli_args.select
    )

    for row, query in enricher.cells(cli_args.query, with_rows=True):

        # Templating?
        if cli_args.query_template is not None:
            query = CUSTOM_FORMATTER.format(cli_args.query_template, value=query)

        loading_bar.print('Searching for "%s"' % query)
        loading_bar.inc("queries")

        if cli_args.items == "tweets":
            iterator = scraper.search_tweets(
                query,
                limit=cli_args.limit,
                include_referenced_tweets=cli_args.include_refs,
                with_meta=True,
            )
        else:
            iterator = scraper.search_users(query, limit=cli_args.limit)

        try:
            for data in iterator:
                loading_bar.update()

                if cli_args.items == "tweets":
                    tweet, meta = data
                    tweet_row = format_tweet_as_csv_row(tweet)
                    addendum = tweet_row + format_meta_row(meta)
                else:
                    addendum = format_user_as_csv_row(data)

                enricher.writerow(row, addendum)

        except TwitterPublicAPIOverCapacityError:
            loading_bar.die('Got an "Over Capacity" error. Shutting down...')



# =============================================================================
# Minet Utils
# =============================================================================
#
# Miscellaneous helper function used throughout the library.
#
import re
import hashlib
import json
import yaml
import time
import string
import functools
import dateparser
from random import uniform

from minet.exceptions import DefinitionInvalidFormatError


def fuzzy_int(value):
    try:
        return int(value)
    except ValueError:
        return int(float(value))


def md5(string):
    h = hashlib.md5()
    h.update(string.encode())
    return h.hexdigest()


DOUBLE_QUOTES_RE = re.compile(r'"')


def fix_ensure_ascii_json_string(s):
    try:
        return json.loads('"%s"' % DOUBLE_QUOTES_RE.sub('\\"', s))
    except json.decoder.JSONDecodeError:
        return s


class RateLimiter(object):
    """
    Naive rate limiter context manager with smooth output.

    Note that it won't work in a multi-threaded environment.

    Args:
        max_per_period (int): Maximum number of calls per period.
        period (float): Duration of a period in seconds. Defaults to 1.0.

    """

    def __init__(self, max_per_period, period=1.0, with_budget=False):
        max_per_second = max_per_period / period
        self.min_interval = 1.0 / max_per_second
        self.max_budget = period / 4
        self.budget = 0.0
        self.last_entry = None
        self.with_budget = with_budget

    def enter(self):
        self.last_entry = time.perf_counter()

    def __enter__(self):
        return self.enter()

    def exit_with_budget(self):
        running_time = time.perf_counter() - self.last_entry

        delta = self.min_interval - running_time

        # Consuming budget
        if delta >= self.budget:
            delta -= self.budget
            self.budget = 0
        else:
            self.budget -= delta
            delta = 0

        # Do we need to sleep?
        if delta > 0:
            time.sleep(delta)
        elif delta < 0:
            self.budget -= delta

        # Clamping budget
        # TODO: this should be improved by a circular buffer of last calls
        self.budget = min(self.budget, self.max_budget)

    def exit(self):
        running_time = time.perf_counter() - self.last_entry

        delta = self.min_interval - running_time

        if delta > 0:
            time.sleep(delta)

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.with_budget:
            return self.exit_with_budget()

        return self.exit()


class RetryableIterator(object):
    """
    Iterator exposing a #.retry method that will make sure the next item
    is the same as the current one.
    """

    def __init__(self, iterator):
        self.iterator = iter(iterator)
        self.current_value = None
        self.retried = False
        self.retries = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.retried:
            self.retried = False
            return self.current_value

        self.retries = 0
        self.current_value = next(self.iterator)
        return self.current_value

    def retry(self):
        self.retries += 1
        self.retried = True


class RateLimitedIterator(object):
    """
    Handy iterator wrapper that will yield its items while respecting a given
    rate limit and that will not sleep needlessly when the iterator is
    finally fully consumed.
    """

    def __init__(self, iterator, max_per_period, period=1.0):
        self.iterator = RetryableIterator(iterator)
        self.rate_limiter = RateLimiter(max_per_period, period)
        self.empty = False

        try:
            self.next_value = next(self.iterator)
        except StopIteration:
            self.next_value = None
            self.empty = True

    @property
    def retries(self):
        return self.iterator.retries

    def retry(self):
        return self.iterator.retry()

    def __iter__(self):
        if self.empty:
            return

        while True:
            self.rate_limiter.enter()

            yield self.next_value

            # NOTE: if the iterator is fully consumed, this will raise StopIteration
            # and skip the exit part that could sleep needlessly
            try:
                self.next_value = next(self.iterator)
            except StopIteration:
                return

            self.rate_limiter.exit()


class RateLimiterState(object):
    def __init__(self, max_per_period, period=1.0):
        max_per_second = max_per_period / period
        self.min_interval = 1.0 / max_per_second
        self.last_entry = None

    def wait_if_needed(self):
        if self.last_entry is None:
            return

        running_time = time.perf_counter() - self.last_entry
        delta = self.min_interval - running_time

        if delta > 0:
            time.sleep(delta)

    def update(self):
        self.last_entry = time.perf_counter()


def rate_limited(max_per_period, period=1.0):
    state = RateLimiterState(max_per_period, period)

    def decorate(fn):
        @functools.wraps(fn)
        def decorated(*args, **kwargs):
            state.wait_if_needed()
            result = fn(*args, **kwargs)
            state.update()

            return result

        return decorated

    return decorate


def rate_limited_from_state(state):
    def decorate(fn):
        @functools.wraps(fn)
        def decorated(*args, **kwargs):
            state.wait_if_needed()
            result = fn(*args, **kwargs)
            state.update()

            return result

        return decorated

    return decorate


def rate_limited_method(attr="rate_limiter_state"):
    def decorate(fn):
        @functools.wraps(fn)
        def decorated(self, *args, **kwargs):
            state = getattr(self, attr)

            if not isinstance(state, RateLimiterState):
                raise ValueError

            state.wait_if_needed()
            result = fn(self, *args, **kwargs)
            state.update()

            return result

        return decorated

    return decorate


class PseudoFStringFormatter(string.Formatter):
    def get_field(self, field_name, args, kwargs):
        result = eval(field_name, None, kwargs)

        return result, None


def load_definition(f, encoding="utf-8"):
    string_path = isinstance(f, str)

    if string_path:
        path = f
        f = open(path, encoding=encoding)
    else:
        path = f.name

    if path.endswith(".json"):
        definition = json.load(f)

    elif path.endswith(".yml") or path.endswith(".yaml"):
        definition = yaml.safe_load(f)

    else:
        raise DefinitionInvalidFormatError

    if string_path:
        f.close()

    return definition


def sleep_with_entropy(seconds, max_random_addendum):
    random_addendum = uniform(0, max_random_addendum)
    time.sleep(seconds + random_addendum)


INTERVALS = [
    ("weeks", 60 * 60 * 24 * 7),  # 60 * 60 * 24 * 7
    ("days", 60 * 60 * 24),  # 60 * 60 * 24
    ("hours", 60 * 60),  # 60 * 60
    ("minutes", 60),
    ("seconds", 1),
]


def prettyprint_integer(n):
    return "{:,}".format(int(n))


def prettyprint_seconds(seconds, granularity=None):
    result = []

    for name, count in INTERVALS:
        value = seconds // count

        if value:
            seconds -= value * count

            if value == 1:
                name = name.rstrip("s")

            result.append("%i %s" % (value, name))

    if not result:
        return "%.2f seconds" % seconds

    if granularity is not None:
        result = result[:granularity]

    return ", ".join(result)


def parse_date(formatted_date, lang="en"):
    if not isinstance(formatted_date, str):
        raise TypeError

    parsed = dateparser.parse(formatted_date, languages=[lang])

    if not parsed:
        return None

    return parsed.isoformat().split(".", 1)[0]


def is_binary_mimetype(m: str) -> bool:
    if m.startswith("text/"):
        return False

    if not m.startswith("application/"):
        return True

    second_part = m.split("/", 1)[-1]

    return not (
        "json" in second_part
        or "html" in second_part
        or "xml" in second_part
        or "yaml" in second_part
        or "yml" in second_part
        or second_part == "x-httpd-php"
    )



# =============================================================================
# Minet Twitter Retweeters CLI Action
# =============================================================================
#
# Logic of the `tw retweeters` action.
#
import casanova
from twitter import TwitterHTTPError
from twitwi import normalize_user, format_user_as_csv_row
from twitwi.constants import USER_PARAMS, USER_FIELDS

from minet.cli.utils import LoadingBar
from minet.twitter import TwitterAPIClient

ITEMS_PER_PAGE = 100


def twitter_retweeters_action(cli_args):
    client = TwitterAPIClient(
        cli_args.access_token,
        cli_args.access_token_secret,
        cli_args.api_key,
        cli_args.api_secret_key,
        api_version="2",
    )

    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        keep=cli_args.select,
        add=USER_FIELDS,
        total=cli_args.total,
    )

    loading_bar = LoadingBar(
        "Retrieving retweeters", total=enricher.total, unit=" users"
    )

    for row, tweet in enricher.cells(cli_args.column, with_rows=True):
        loading_bar.inc("tweets")
        kwargs = {"max_results": ITEMS_PER_PAGE, "params": USER_PARAMS}

        while True:
            try:
                result = client.call(["tweets", tweet, "retweeted_by"], **kwargs)
            except TwitterHTTPError as e:
                loading_bar.inc("errors")

                if e.e.code == 404:
                    enricher.writerow(row)
                else:
                    raise e

                continue

            if "data" not in "result" and result["meta"]["result_count"] == 0:
                break

            for user in result["data"]:
                user = normalize_user(user, v2=True)
                user_row = format_user_as_csv_row(user)
                enricher.writerow(row, user_row)
                loading_bar.update()

            if "next_token" in result["meta"]:
                kwargs["pagination_token"] = result["meta"]["next_token"]
            else:
                break



# =============================================================================
# Minet Instagram API Scraper
# =============================================================================
#
# Instagram public API "scraper".
#
import json
from urllib.parse import quote
from ebbe import getpath

from minet.constants import COOKIE_BROWSERS
from minet.web import create_pool, request_json, grab_cookies
from minet.instagram.constants import (
    INSTAGRAM_URL,
    INSTAGRAM_PUBLIC_API_DEFAULT_TIMEOUT,
)
from minet.instagram.exceptions import (
    InstagramPublicAPIInvalidResponseError,
    InstagramInvalidCookieError,
)

INSTAGRAM_GRAPHQL_ENDPOINT = "https://www.instagram.com/graphql/query/"
INSTAGRAM_HASHTAG_QUERY_HASH = "9b498c08113f1e09617a1703c22b2f32"


def forge_hashtag_search_url(name, cursor=None, count=50):
    params = {"tag_name": name, "first": count}

    if cursor is not None:
        params["after"] = cursor

    url = INSTAGRAM_GRAPHQL_ENDPOINT + "?query_hash=%s&variables=%s" % (
        INSTAGRAM_HASHTAG_QUERY_HASH,
        quote(json.dumps(params)),
    )

    return url


class InstagramAPIScraper(object):
    def __init__(self, cookie="firefox"):
        self.pool = create_pool(timeout=INSTAGRAM_PUBLIC_API_DEFAULT_TIMEOUT)

        if cookie in COOKIE_BROWSERS:
            get_cookie_for_url = grab_cookies(cookie)
            cookie = get_cookie_for_url(INSTAGRAM_URL)

        if not cookie:
            raise InstagramInvalidCookieError

        self.cookie = cookie

    def request_json(self, url):
        err, response, data = request_json(
            url, pool=self.pool, spoof_ua=True, headers={"Cookie": self.cookie}
        )

        if err:
            raise err

        if response.status >= 400:
            raise InstagramPublicAPIInvalidResponseError

        return data

    def search_hashtag(self, name):
        name = name.lstrip("#")
        cursor = None

        while True:
            url = forge_hashtag_search_url(name, cursor=cursor)
            print(url, cursor)

            data = self.request_json(url)

            data = getpath(data, ["data", "hashtag", "edge_hashtag_to_media"])
            edges = data.get("edges")

            for edge in edges:
                yield edge["node"]["shortcode"]

            print("Found %i posts" % len(edges))

            has_next_page = getpath(data, ["page_info", "has_next_page"])

            if not has_next_page:
                break

            cursor = getpath(data, ["page_info", "end_cursor"])



# =============================================================================
# Minet CrowdTangle Posts By Id CLI Action
# =============================================================================
#
# Logic of the `ct posts-by-id` action.
#
import casanova
from ural import ensure_protocol
from ural.facebook import is_facebook_post_url

import minet.facebook as facebook
from minet.cli.utils import LoadingBar
from minet.crowdtangle.constants import CROWDTANGLE_POST_CSV_HEADERS
from minet.crowdtangle.client import CrowdTangleAPIClient
from minet.crowdtangle.exceptions import CrowdTangleInvalidTokenError


def crowdtangle_posts_by_id_action(cli_args):

    client = CrowdTangleAPIClient(cli_args.token, rate_limit=cli_args.rate_limit)

    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        keep=cli_args.select,
        add=CROWDTANGLE_POST_CSV_HEADERS,
    )

    loading_bar = LoadingBar(desc="Retrieving posts", total=cli_args.total, unit="post")

    try:
        for row, url in enricher.cells(cli_args.column, with_rows=True):
            loading_bar.update()

            url = url.strip()

            if not url:
                enricher.writerow(row)
                continue

            url = ensure_protocol(url)

            if not is_facebook_post_url(url):
                enricher.writerow(row)
                continue

            post_id = facebook.post_id_from_url(url)

            if post_id is None:
                enricher.writerow(row)
                continue

            post = client.post(post_id)
            enricher.writerow(row, post)

    except CrowdTangleInvalidTokenError:
        loading_bar.die(
            [
                "Your API token is invalid.",
                "Check that you indicated a valid one using the `--token` argument.",
            ]
        )



# =============================================================================
# Minet Twitter User Tweets CLI Action
# =============================================================================
#
# Logic of the `tw user-tweets` action.
#
import casanova
from datetime import datetime
from twitwi import normalize_tweet, normalize_tweets_payload_v2, format_tweet_as_csv_row
from twitwi.constants import TWEET_FIELDS, TWEET_EXPANSIONS, TWEET_PARAMS
from twitter import TwitterHTTPError

from minet.cli.utils import LoadingBar
from minet.twitter.constants import TWITTER_API_MAX_STATUSES_COUNT
from minet.twitter import TwitterAPIClient
from minet.cli.twitter.utils import is_not_user_id, is_probably_not_user_screen_name

ITEMS_PER_PAGE = 100


def twitter_user_tweets_action(cli_args):

    client = TwitterAPIClient(
        cli_args.access_token,
        cli_args.access_token_secret,
        cli_args.api_key,
        cli_args.api_secret_key,
        api_version="2" if cli_args.v2 else "1.1",
    )

    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        keep=cli_args.select,
        add=TWEET_FIELDS,
        total=cli_args.total,
    )

    loading_bar = LoadingBar("Retrieving tweets", total=enricher.total, unit="user")

    for row, user in enricher.cells(cli_args.column, with_rows=True):
        max_id = None

        loading_bar.update_stats(user=user)

        min_date_reached = False

        kwargs = {}

        while True:
            if cli_args.v2:
                if is_not_user_id(user):
                    loading_bar.die(
                        "The column given as argument doesn't contain user ids, you have probably given user screen names as argument instead. With --api-v2, you can only use user ids to retrieve followers."
                    )

                kwargs["max_results"] = ITEMS_PER_PAGE
                kwargs["params"] = TWEET_PARAMS
                kwargs["expansions"] = ",".join(TWEET_EXPANSIONS)

                if cli_args.exclude_retweets:
                    kwargs["exclude"] = "retweets"
                if cli_args.min_date:
                    utc_dt = datetime.fromtimestamp(cli_args.min_date).strftime(
                        "%Y-%m-%d"
                    )
                    kwargs["start_time"] = utc_dt + "T00:00:00Z"

                try:
                    tweets = client.call(["users", user, "tweets"], **kwargs)
                except TwitterHTTPError as e:
                    loading_bar.print(e)
                    loading_bar.inc("errors")

                    if e.e.code == 404:
                        loading_bar.print('Could not find user "%s"' % user)
                    else:
                        loading_bar.print(
                            'An error happened when attempting to retrieve tweets from "%s" (HTTP status %i)'
                            % (user, e.e.code)
                        )

                    break

                try:
                    normalized_tweets = normalize_tweets_payload_v2(
                        tweets, collection_source="api"
                    )
                except TypeError:
                    loading_bar.print(
                        'An error happened when attempting to retrieve tweets from "%s"'
                        % user
                    )
                    loading_bar.print('details: "%s"' % tweets["errors"][0]["detail"])
                    break

                for normalized_tweet in normalized_tweets:
                    addendum = format_tweet_as_csv_row(normalized_tweet)
                    enricher.writerow(row, addendum)

                if "next_token" in tweets["meta"]:
                    kwargs["pagination_token"] = tweets["meta"]["next_token"]
                else:
                    break

            else:
                if cli_args.ids:
                    if is_not_user_id(user):
                        loading_bar.die(
                            "The column given as argument doesn't contain user ids, you have probably given user screen names as argument instead. \nTry removing --ids from the command."
                        )

                    kwargs = {"user_id": user}

                else:
                    if is_probably_not_user_screen_name(user):
                        loading_bar.die(
                            "The column given as argument probably doesn't contain user screen names, you have probably given user ids as argument instead. \nTry adding --ids to the command."
                        )
                        # force flag to add

                    kwargs = {"screen_name": user}

                kwargs["include_rts"] = not cli_args.exclude_retweets
                kwargs["count"] = TWITTER_API_MAX_STATUSES_COUNT
                kwargs["tweet_mode"] = "extended"

                if max_id is not None:
                    kwargs["max_id"] = max_id

                loading_bar.inc("calls")

                try:
                    tweets = client.call(["statuses", "user_timeline"], **kwargs)
                except TwitterHTTPError as e:
                    loading_bar.inc("errors")

                    if e.e.code == 404:
                        loading_bar.print('Could not find user "%s"' % user)
                    else:
                        loading_bar.print(
                            'An error happened when attempting to retrieve tweets from "%s" (HTTP status %i)'
                            % (user, e.e.code)
                        )

                    break

                if not tweets:
                    break

                max_id = min(int(tweet["id_str"]) for tweet in tweets) - 1

                for tweet in tweets:
                    tweet = normalize_tweet(tweet, collection_source="api")
                    addendum = format_tweet_as_csv_row(tweet)

                    if cli_args.min_date:
                        if int(addendum[1]) < cli_args.min_date:
                            min_date_reached = True
                            break

                    enricher.writerow(row, addendum)

            loading_bar.inc("tweets", len(tweets))

            if min_date_reached:
                break

        loading_bar.update()



# =============================================================================
# Minet Twitter Tweet Count CLI Action
# =============================================================================
#
# Logic of the `tw tweet-count` action.
#
import casanova
from twitter import TwitterHTTPError

from minet.cli.utils import LoadingBar
from minet.cli.twitter.utils import validate_query_boundaries
from minet.twitter import TwitterAPIClient

ITEMS_PER_PAGE = 100

COUNT_FIELDS = ["tweet_count"]
GRANULARIZED_COUNT_FIELDS = ["start_time", "end_time", "tweet_count"]


def twitter_tweet_count_action(cli_args):
    validate_query_boundaries(cli_args)

    client = TwitterAPIClient(
        cli_args.access_token,
        cli_args.access_token_secret,
        cli_args.api_key,
        cli_args.api_secret_key,
        api_version="2",
    )

    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        keep=cli_args.select,
        add=GRANULARIZED_COUNT_FIELDS
        if cli_args.granularity is not None
        else COUNT_FIELDS,
        total=cli_args.total,
    )

    loading_bar = LoadingBar(
        "Counting tweets", total=enricher.total, unit="query", unit_plural="queries"
    )

    for row, query in enricher.cells(cli_args.query, with_rows=True):

        kwargs = {"query": query}

        loading_bar.print('Counting tweets for "%s"' % query)

        # Because we are greedy, we want stuff from the beginning of Twitter
        if cli_args.academic and not cli_args.start_time:
            kwargs["start_time"] = "2006-03-21T00:00:00Z"

        if cli_args.start_time:
            kwargs["start_time"] = cli_args.start_time
        if cli_args.end_time:
            kwargs["end_time"] = cli_args.end_time
        if cli_args.since_id:
            kwargs["since_id"] = cli_args.since_id
        if cli_args.until_id:
            kwargs["until_id"] = cli_args.until_id

        kwargs["granularity"] = cli_args.granularity or "day"

        route = (
            ["tweets", "counts", "all"]
            if cli_args.academic
            else ["tweets", "counts", "recent"]
        )

        total_count = 0

        while True:
            try:
                result = client.call(route, **kwargs)
                loading_bar.inc("calls")
            except TwitterHTTPError as e:
                loading_bar.inc("errors")

                if e.e.code == 404:
                    enricher.writerow(row)
                else:
                    raise e

                continue

            for count in result["data"]:
                total_count += count["tweet_count"]

                if cli_args.granularity is not None:
                    addendum = [count["start"], count["end"], count["tweet_count"]]
                    enricher.writerow(row, addendum)

            if "next_token" in result["meta"]:
                kwargs["next_token"] = result["meta"]["next_token"]
            else:
                break

        loading_bar.update()

        if cli_args.granularity is None:
            enricher.writerow(row, [total_count])



# =============================================================================
# Minet YouTube API Client
# =============================================================================
#
# A handy API client used by the CLI actions.
#
import time
from ebbe import as_chunks
from collections import deque
from urllib.parse import quote
from ebbe import getpath

from minet.web import create_pool, request_json
from minet.youtube.utils import (
    ensure_video_id,
    ensure_channel_id,
    get_channel_main_playlist_id,
    seconds_to_midnight_pacific_time,
)
from minet.youtube.constants import (
    YOUTUBE_API_BASE_URL,
    YOUTUBE_API_MAX_VIDEOS_PER_CALL,
    YOUTUBE_API_MAX_COMMENTS_PER_CALL,
    YOUTUBE_API_DEFAULT_SEARCH_ORDER,
    YOUTUBE_API_SEARCH_ORDERS,
)
from minet.youtube.exceptions import (
    YouTubeDisabledComments,
    YouTubeVideoNotFound,
    YouTubeInvalidAPIKeyError,
    YouTubeInvalidAPICall,
    YouTubeInvalidVideoTarget,
    YouTubeInvalidChannelTarget,
)
from minet.youtube.formatters import (
    format_video,
    format_video_snippet,
    format_comment,
    format_reply,
    format_playlist_item_snippet,
)
from minet.youtube.scrapers import scrape_channel_id


def forge_playlist_videos_url(key, playlist_id, token=None):
    data = {
        "base": YOUTUBE_API_BASE_URL,
        "playlist_id": playlist_id,
        "key": key,
        "count": YOUTUBE_API_MAX_VIDEOS_PER_CALL,
    }

    url = (
        "%(base)s/playlistItems?part=snippet&maxResults=%(count)i&playlistId=%(playlist_id)s&key=%(key)s"
        % data
    )

    if token is not None:
        url += "&pageToken=%s" % token

    return url


def forge_videos_url(key, ids):
    data = {"base": YOUTUBE_API_BASE_URL, "ids": ",".join(ids), "key": key}

    return (
        "%(base)s/videos?id=%(ids)s&key=%(key)s&part=snippet,statistics,contentDetails"
        % data
    )


def forge_search_url(key, query, order=YOUTUBE_API_DEFAULT_SEARCH_ORDER, token=None):
    data = {
        "base": YOUTUBE_API_BASE_URL,
        "order": order,
        "query": quote(query),
        "key": key,
        "count": YOUTUBE_API_MAX_VIDEOS_PER_CALL,
    }

    url = (
        "%(base)s/search?part=snippet&maxResults=%(count)i&q=%(query)s&type=video&order=%(order)s&key=%(key)s"
        % data
    )

    if token is not None:
        url += "&pageToken=%s" % token

    return url


def forge_comments_url(key, video_id, token=None):
    data = {
        "base": YOUTUBE_API_BASE_URL,
        "key": key,
        "count": YOUTUBE_API_MAX_COMMENTS_PER_CALL,
        "video_id": video_id,
    }

    url = (
        "%(base)s/commentThreads?videoId=%(video_id)s&key=%(key)s&part=snippet,replies&maxResults=%(count)s"
        % data
    )

    if token is not None:
        url += "&pageToken=%s" % token

    return url


def forge_replies_url(key, comment_id, token=None):
    data = {
        "base": YOUTUBE_API_BASE_URL,
        "key": key,
        "comment_id": comment_id,
        "count": YOUTUBE_API_MAX_COMMENTS_PER_CALL,
    }

    url = (
        "%(base)s/comments?part=snippet&parentId=%(comment_id)s&key=%(key)s&maxResults=%(count)s"
        % data
    )

    if token is not None:
        url += "&pageToken=%s" % token

    return url


class YouTubeAPIClient(object):
    def __init__(self, key, before_sleep_until_midnight=None):
        self.key = key
        self.pool = create_pool()
        self.before_sleep = before_sleep_until_midnight

    def request_json(self, url):
        err, response, data = request_json(url, pool=self.pool)

        if err:
            raise err

        if response.status == 403:
            if (
                data is not None
                and getpath(data, ["error", "errors", 0, "reason"], "")
                == "commentsDisabled"
            ):
                raise YouTubeDisabledComments

            sleep_time = seconds_to_midnight_pacific_time() + 10

            if callable(self.before_sleep):
                self.before_sleep(sleep_time)

            time.sleep(sleep_time)

            return self.request_json(url)

        if response.status == 404:
            raise YouTubeVideoNotFound

        if response.status >= 400:
            if data is not None and "API key not valid" in getpath(
                data, ["error", "message"], ""
            ):
                raise YouTubeInvalidAPIKeyError

            raise YouTubeInvalidAPICall(url, response.status, data)

        return data

    def videos(self, videos, key=None, raw=False):

        # TODO: we could chunk per not None
        for group in as_chunks(YOUTUBE_API_MAX_VIDEOS_PER_CALL, videos):
            group_data = []

            for item in group:
                target = key(item) if key is not None else item
                video_id = ensure_video_id(target)
                group_data.append((video_id, item))

            ids = [video_id for video_id, _ in group_data if video_id is not None]

            url = forge_videos_url(self.key, ids)

            result = self.request_json(url)

            indexed_result = {}

            for item in result["items"]:
                video_id = item["id"]

                if not raw:
                    item = format_video(item)

                indexed_result[video_id] = item

            for video_id, item in group_data:
                yield item, indexed_result.get(video_id)

    def search(self, query, order=YOUTUBE_API_DEFAULT_SEARCH_ORDER, raw=False):
        if order not in YOUTUBE_API_SEARCH_ORDERS:
            raise TypeError('unkown search order "%s"' % order)

        def generator():
            token = None

            while True:
                url = forge_search_url(self.key, query, order=order, token=token)

                result = self.request_json(url)

                token = result.get("nextPageToken")

                for item in result["items"]:
                    if not raw:
                        item = format_video_snippet(item)

                    yield item

                if token is None or len(result["items"]) == 0:
                    break

        return generator()

    def comments(self, video_target, raw=False, full_replies=True):
        video_id = ensure_video_id(video_target)

        if video_id is None:
            raise YouTubeInvalidVideoTarget

        def generator():
            starting_url = forge_comments_url(self.key, video_id)

            queue = deque([(False, video_id, starting_url)])

            while len(queue) != 0:
                is_reply, item_id, url = queue.popleft()

                try:
                    result = self.request_json(url)

                except (YouTubeDisabledComments, YouTubeVideoNotFound):
                    return

                for item in result["items"]:
                    comment_id = item["id"]
                    replies = getpath(item, ["replies", "comments"], [])
                    total_reply_count = getpath(item, ["snippet", "totalReplyCount"], 0)

                    if not raw:
                        item = (
                            format_comment(item)
                            if not is_reply
                            else format_reply(item, video_id=video_id)
                        )

                    yield item

                    if is_reply:
                        continue

                    # Getting replies
                    if not full_replies or len(replies) >= total_reply_count:
                        for reply in replies:
                            if not raw:
                                reply = format_reply(reply)

                            yield reply
                    elif total_reply_count > 0:
                        replies_url = forge_replies_url(self.key, comment_id)

                        queue.append((True, comment_id, replies_url))

                # Next page
                token = result.get("nextPageToken")

                if token is not None and len(result["items"]) != 0:
                    forge = forge_replies_url if is_reply else forge_comments_url

                    next_url = forge(self.key, item_id, token=token)

                    queue.append((is_reply, item_id, next_url))

        return generator()

    def channel_videos(self, channel_target):

        should_scrape, channel_id = ensure_channel_id(channel_target)

        if should_scrape:
            channel_id = scrape_channel_id(channel_target)

        if channel_id is None:
            raise YouTubeInvalidChannelTarget

        playlist_id = get_channel_main_playlist_id(channel_id)

        def generator():
            token = None

            while True:
                url = forge_playlist_videos_url(self.key, playlist_id, token=token)

                result = self.request_json(url)

                token = result.get("nextPageToken")

                for item in result["items"]:
                    item = format_playlist_item_snippet(item)

                    yield item

                if token is None or len(result["items"]) == 0:
                    break

        return generator()



# =============================================================================
# Minet Url Extract CLI Action
# =============================================================================
#
# Logic of the `url-extract` action.
#
import casanova
from ural import urls_from_text, urls_from_html
from urllib.parse import urljoin

from minet.cli.utils import LoadingBar

REPORT_HEADERS = ["extracted_url"]

EXTRACTORS = {"html": urls_from_html, "text": urls_from_text}


def url_extract_action(cli_args):
    enricher = casanova.enricher(
        cli_args.file, cli_args.output, add=REPORT_HEADERS, keep=cli_args.select
    )

    extract = EXTRACTORS[getattr(cli_args, "from")]

    loading_bar = LoadingBar(desc="Extracting", unit="row", total=cli_args.total)

    for row, content in enricher.cells(cli_args.column, with_rows=True):
        loading_bar.update()

        content = content.strip()

        if not content:
            continue

        for url in extract(content):
            if cli_args.base_url is not None:
                url = urljoin(cli_args.base_url, url)

            enricher.writerow(row, [url])



# =============================================================================
# Minet Facebook Utils
# =============================================================================
#
# Miscellaneous helpers used by the minet.facebook namespace.
#
from http.cookies import SimpleCookie

from minet.web import grab_cookies
from minet.constants import COOKIE_BROWSERS
from minet.facebook.constants import FACEBOOK_URL


def fix_cookie(cookie_string):
    cookie = SimpleCookie()
    cookie.load(cookie_string)

    # NOTE: those cookie items can rat you out
    try:
        del cookie["m_pixel_ratio"]
        del cookie["wd"]
    except KeyError:
        pass

    cookie["locale"] = "en_US"

    return "; ".join(key + "=" + morsel.coded_value for key, morsel in cookie.items())


def grab_facebook_cookie(source):
    if source in COOKIE_BROWSERS:
        get_cookie_for_url = grab_cookies(source)

        if get_cookie_for_url is None:
            return None

        cookie = get_cookie_for_url(FACEBOOK_URL + "/")

    else:
        cookie = source.strip()

    if not cookie or "c_user=" not in cookie.lower():
        return None

    return fix_cookie(cookie)



# =============================================================================
# Minet CrowdTangle Constants
# =============================================================================
#
# General constants used throughout the CrowdTangle functions.
#
from urllib3 import Timeout

CROWDTANGLE_DEFAULT_RATE_LIMIT = 6  # Number of hits per minute
CROWDTANGLE_LINKS_DEFAULT_RATE_LIMIT = 2

CROWDTANGLE_DEFAULT_TIMEOUT = Timeout(connect=10, read=60 * 5)

CROWDTANGLE_POST_TYPES = [
    "episode",
    "extra_clip",
    "link",
    "live_video",
    "live_video_complete",
    "live_video_scheduled",
    "native_video",
    "photo",
    "status",
    "trailer",
    "tweet",
    "video",
    "vine",
    "youtube",
]

CROWDTANGLE_SORT_TYPES = [
    "date",
    "interaction_rate",
    "overperforming",
    "total_interactions",
    "underperforming",
]

CROWDTANGLE_SUMMARY_SORT_TYPES = {"date", "subscriber_count", "total_interactions"}

CROWDTANGLE_SUMMARY_DEFAULT_SORT_TYPE = "date"

CROWDTANGLE_SEARCH_FIELDS = {
    "text_fields_and_image_text",
    "include_query_strings",
    "text_fields_only",
    "account_name_only",
    "image_text_only",
}

CROWDTANGLE_DEFAULT_SEARCH_FIELD = "text_fields_and_image_text"

CROWDTANGLE_STATISTICS = [
    "like",
    "share",
    "favorite",
    "comment",
    "love",
    "wow",
    "haha",
    "sad",
    "angry",
    "thankful",
    "care",
]

CROWDTANGLE_FULL_STATISTICS = STATISTICS = [
    ("loveCount", "love_count"),
    ("wowCount", "wow_count"),
    ("thankfulCount", "thankful_count"),
    ("interactionRate", "interaction_rate"),
    ("likeCount", "like_count"),
    ("hahaCount", "haha_count"),
    ("commentCount", "comment_count"),
    ("shareCount", "share_count"),
    ("sadCount", "sad_count"),
    ("angryCount", "angry_count"),
    ("postCount", "post_count"),
    ("totalInteractionCount", "total_interaction_count"),
    ("totalVideoTimeMS", "total_video_time_ms"),
    ("threePlusMinuteVideoCount", "three_plus_minute_video_count"),
]

CROWDTANGLE_PLATFORMS = {"facebook", "instagram", "reddit"}

CROWDTANGLE_REACTION_TYPES = [
    "angry",
    "comment",
    "haha",
    "like",
    "love",
    "sad",
    "share",
    "thankful",
    "wow",
]

CROWDTANGLE_POST_CSV_HEADERS = [
    "ct_id",
    "id",
    "platform",
    "type",
    "title",
    "caption",
    "message",
    "description",
    "date",
    "datetime",
    "updated",
    "link",
    "post_url",
    "score",
    "video_length_ms",
    "live_video_status",
]

for name in CROWDTANGLE_STATISTICS:
    CROWDTANGLE_POST_CSV_HEADERS.append("actual_%s_count" % name)
    CROWDTANGLE_POST_CSV_HEADERS.append("expected_%s_count" % name)

CROWDTANGLE_ACCOUNT_CSV_HEADERS = [
    "account_ct_id",
    "account_id",
    "account_platform",
    "account_name",
    "account_handle",
    "account_profile_image",
    "account_subscriber_count",
    "account_url",
    "account_verified",
    "account_type",
    "account_page_admin_top_country",
]

CROWDTANGLE_MEDIA_CSV_HEADERS = ["links", "expanded_links", "media"]

CROWDTANGLE_POST_CSV_HEADERS += CROWDTANGLE_ACCOUNT_CSV_HEADERS
CROWDTANGLE_POST_CSV_HEADERS += CROWDTANGLE_MEDIA_CSV_HEADERS

CROWDTANGLE_POST_CSV_HEADERS_WITH_LINK = ["url"] + CROWDTANGLE_POST_CSV_HEADERS

CROWDTANGLE_SUMMARY_CSV_HEADERS = ["%s_count" % t for t in CROWDTANGLE_REACTION_TYPES]

CROWDTANGLE_LEADERBOARD_CSV_HEADERS = [
    "ct_id",
    "name",
    "handle",
    "profile_image",
    "subscriber_count",
    "url",
    "verified",
    "initial_subscriber_count",
    "final_subscriber_count",
    "subscriber_data_notes",
]

for _, substitute_key in CROWDTANGLE_FULL_STATISTICS:
    CROWDTANGLE_LEADERBOARD_CSV_HEADERS.append(substitute_key)

CROWDTANGLE_LEADERBOARD_CSV_HEADERS_WITH_BREAKDOWN = list(
    CROWDTANGLE_LEADERBOARD_CSV_HEADERS
)

for post_type in CROWDTANGLE_POST_TYPES:
    for _, substitute_key in CROWDTANGLE_FULL_STATISTICS:
        CROWDTANGLE_LEADERBOARD_CSV_HEADERS_WITH_BREAKDOWN.append(
            "%s_%s" % (post_type, substitute_key)
        )

CROWDTANGLE_LIST_CSV_HEADERS = ["id", "title", "type"]



# =============================================================================
# Minet CrowdTangle Leaderboard CLI Action
# =============================================================================
#
# Logic of the `ct leaderboard` action.
#
from minet.crowdtangle.constants import (
    CROWDTANGLE_LEADERBOARD_CSV_HEADERS,
    CROWDTANGLE_LEADERBOARD_CSV_HEADERS_WITH_BREAKDOWN,
)
from minet.cli.crowdtangle.utils import make_paginated_action


def select_csv_headers(cli_args):
    if cli_args.breakdown:
        return CROWDTANGLE_LEADERBOARD_CSV_HEADERS_WITH_BREAKDOWN

    return CROWDTANGLE_LEADERBOARD_CSV_HEADERS


crowdtangle_leaderboard_action = make_paginated_action(
    method_name="leaderboard", item_name="accounts", csv_headers=select_csv_headers
)



# =============================================================================
# Minet Hyphe Constants
# =============================================================================
#
# Bunch of Hyphe-related constants.
#

WEBENTITY_STATUSES = set(["IN", "OUT", "UNDECIDED", "DISCOVERED"])

DEFAULT_PAGINATION_COUNT = 100

WEBENTITY_CSV_HEADERS = [
    "id",
    "name",
    "status",
    "pages",
    "homepage",
    "prefixes",
    "degree",
    "undirected_degree",
    "indegree",
    "outdegree",
]

PAGE_CSV_HEADERS = [
    "url",
    "lru",
    "webentity",
    "webentity_status",
    "status",
    "crawled",
    "encoding",
    "content_type",
    "crawl_timestamp",
    "crawl_datetime",
    "size",
    "error",
]



# =============================================================================
# Minet Twitter CLI Action
# =============================================================================
#
# Logic of the `tw` action.
#
from minet.cli.utils import die


def check_credentials(cli_args):

    # Credentials are required to be able to access the API
    if (
        not cli_args.api_key
        or not cli_args.api_secret_key
        or not cli_args.access_token
        or not cli_args.access_token_secret
    ):
        die(
            [
                "Full credentials are required to access Twitter API.",
                "You can provide them using various CLI arguments:",
                "    --api-key",
                "    --api-secret-key",
                "    --access-token",
                "    --access-token-secret",
            ]
        )


def twitter_action(cli_args):
    if cli_args.tw_action == "scrape":
        from minet.cli.twitter.scrape import twitter_scrape_action

        twitter_scrape_action(cli_args)

    else:
        check_credentials(cli_args)

        if cli_args.tw_action == "friends":
            from minet.cli.twitter.friends import twitter_friends_action

            twitter_friends_action(cli_args)

        elif cli_args.tw_action == "followers":
            from minet.cli.twitter.followers import twitter_followers_action

            twitter_followers_action(cli_args)

        elif cli_args.tw_action == "users":
            from minet.cli.twitter.users import twitter_users_action

            twitter_users_action(cli_args)

        elif cli_args.tw_action == "user-tweets":
            from minet.cli.twitter.user_tweets import twitter_user_tweets_action

            twitter_user_tweets_action(cli_args)

        elif cli_args.tw_action == "tweet-search":
            from minet.cli.twitter.tweet_search import twitter_tweet_search_action

            twitter_tweet_search_action(cli_args)

        elif cli_args.tw_action == "tweet-count":
            from minet.cli.twitter.tweet_count import twitter_tweet_count_action

            twitter_tweet_count_action(cli_args)

        elif cli_args.tw_action == "tweets":
            from minet.cli.twitter.tweets import twitter_tweets_action

            twitter_tweets_action(cli_args)

        elif cli_args.tw_action == "attrition":
            from minet.cli.twitter.attrition import twitter_attrition_action

            twitter_attrition_action(cli_args)

        elif cli_args.tw_action == "user-search":
            from minet.cli.twitter.user_search import twitter_user_search_action

            twitter_user_search_action(cli_args)

        elif cli_args.tw_action == "retweeters":
            from minet.cli.twitter.retweeters import twitter_retweeters_action

            twitter_retweeters_action(cli_args)

        else:
            raise TypeError('unkown tw_action "%s"' % cli_args.tw_action)
